import json

from jinja2 import Environment, Template

from .titanium_field import TitaniumField

template_string = """/**
 * @file {{ package_name }}Protobuf.h
 * @brief Auto-generated header file generated.
 */

#ifndef {{ package_name | upper }}_PROTO_H
#define {{ package_name | upper }}_PROTO_H

#include "stdint.h"
#include "string.h"
#include "jsmn.h"

class {{ package_name }}Protobuf {
public:
    {{ package_name }}Protobuf() = default;
    ~{{ package_name }}Protobuf() = default;
{% for field in fields %}
{%- if field.is_array %}
    static constexpr uint16_t {{ field.defined_size }} = {{ field.size }};
{%- endif %}
{%- endfor %}
{% for field in fields %}
{%- if field.is_array %}
    const {{ field.c_type_name }}* Get{{ field.capitalized_name }}(void) const { return this->{{ field.internal_name }}; }
{%- else %}
    {{ field.c_type_name }} Get{{ field.capitalized_name }}(void) const { return this->{{ field.internal_name }}; }
{%- endif %}
{%- endfor %}

    int16_t GetSerializedSize(void) const {
        return ({{ proto.serialized_size }});
    }

    int16_t GetMaxSize(void) const {
        return ({{ proto.maximum_size }});
    }

    static int16_t GetStaticMaxSize(void) {
        return ({{ proto.static_maximum_size }});
    }
{% for field in fields -%}
{%- if field.is_array %}
    int8_t Update{{ field.capitalized_name }}(const {{ field.c_type_name }}* value) {
        if (value == nullptr) {
            return -1;
        }

        size_t value_length = strlen(value) + 1;

        if ((value_length == 0) || {{ field.defined_size }} == 0) {
            return -2;
        }

        if (value_length > {{ field.defined_size }}) {
            return -3;
        }

        memset(this->{{ field.internal_name }}, 0, {{ field.defined_size }});
        memcpy(this->{{ field.internal_name }}, value, value_length);

        return 0;
    }

    int8_t Update{{ field.capitalized_name }}(const {{ field.c_type_name }}* value, uint16_t string_size) {
        if (value == nullptr) {
            return -1;
        }

        if ({{ field.defined_size }} == 0) {
            return -2;
        }

        if (string_size > {{ field.defined_size }}) {
            return -3;
        }

        memset(this->{{ field.internal_name }}, 0, {{ field.defined_size }});
        memcpy(this->{{ field.internal_name }}, value, string_size);

        return 0;
    }
{% else %}
    int8_t Update{{ field.capitalized_name }}({{  field.c_type_name }} value) {
        this->{{ field.internal_name }} = value;
        return 0;
    }
{% endif -%}
{% endfor %}
    int16_t Serialize(char* out_buffer, uint16_t out_buffer_size) const {
        if (out_buffer == nullptr) {
            return 0;
        }

        uint16_t serialized_size = {{ proto.serialized_size }};

        if (out_buffer_size < serialized_size) {
            return 0;
        }

        uint16_t offset = 0;
{% for field in fields -%}
{%- if field.is_array %}
        memcpy(&out_buffer[offset], this->{{ field.internal_name }}, strlen(this->{{ field.internal_name }}) + 1);
{%- if not loop.last %}
        offset += strlen(this->{{ field.internal_name }}) + 1;
{%- endif %}
{%- else %}
        memcpy(&out_buffer[offset], &this->{{ field.internal_name }}, sizeof(this->{{ field.internal_name }}));
{%- if not loop.last %}
        offset += sizeof(this->{{ field.internal_name }});
{%- endif %}
{%- endif %}
{%- endfor %}

        return serialized_size;
    }

    int8_t DeSerialize(const char* in_buffer, uint16_t in_buffer_size) {
        if (in_buffer == nullptr) {
            return -1;
        }

        uint16_t deserialized_min_size = {{ proto.minimum_size }};
        uint16_t deserialized_max_size = {{ proto.maximum_size }};

        if ((in_buffer_size < deserialized_min_size) || (in_buffer_size > deserialized_max_size)) {
            return -3;
        }
{% for field in fields -%}
{%- if field.is_array %}
        memset(this->{{ field.internal_name }}, 0, {{ field.defined_size }});
{%- endif %}
{%- endfor %}

        uint16_t offset = 0;

{%- for field in fields %}
{%- if field.is_array %}
        memcpy(this->{{ field.internal_name }}, &in_buffer[offset], strlen(&in_buffer[offset]) + 1);
{%- if not loop.last %}
        offset += strlen(&in_buffer[offset]) + 1;
{%- endif %}
{%- else %}
        memcpy(&this->{{ field.internal_name }}, &in_buffer[offset], sizeof(this->{{ field.internal_name }}));
{%- if not loop.last %}
        offset += sizeof(this->{{ field.internal_name }});
{%- endif %}
{%- endif %}
{%- endfor %}

        return 0;
    }

    int32_t SerializeJson(char* out_buffer, uint16_t out_buffer_size) {
        uint32_t response_length = 0;

        do {
            if (out_buffer == nullptr) {
                break;
            }

            uint16_t serialized_size = {{ proto.serialized_size }};

            if (out_buffer_size < serialized_size) {
                return 0;
            }

            response_length = snprintf(out_buffer, out_buffer_size,
                                       this->_json_string,
{%- for field in fields %}
                                       this->{{ field.internal_name }}{% if not loop.last -%},{%- else %});{% endif -%}
{%- endfor %}
        } while (0);

        return response_length;
    }

    int8_t DeSerializeJson(const char* in_buffer, uint16_t in_buffer_size) {
        auto result = -1;
        jsmn_parser parser;
        jsmntok_t tokens[this->_NUM_TOKENS];

        jsmn_init(&parser);

        do {
            if (in_buffer == nullptr) {
                break;
            }

            auto num_tokens = jsmn_parse(&parser, in_buffer, strlen(in_buffer), tokens, this->_NUM_TOKENS);

            if (num_tokens != this->_NUM_TOKENS) {
                break;
            }

            jsmntok_t key{};
            jsmntok_t value{};
            uint16_t token_length = 0;

{%- for field in fields %}

            key   = tokens[this->{{ field.internal_name  | upper }}_TOKEN_ID];
            value = tokens[this->{{ field.internal_name  | upper }}_TOKEN_ID + 1];
            token_length = key.end - key.start;

            if (strncmp(in_buffer + key.start, this->{{ field.internal_name | upper }}_TOKEN_NAME, token_length) != 0) {
                break;
            }
{%- if field.is_array %}

            this->Update{{ field.capitalized_name }}(in_buffer + value.start, value.end - value.start);
{%- else %}

            this->Update{{ field.capitalized_name }}(atoi(in_buffer + value.start));
{%- endif %}
{%- endfor %}

            result = 0;

        } while(0);

        return result;
    }

private:
{%- for field in fields %}
    {{ field.c_type_name }} {{ field.internal_name }}{% if field.is_array %}[{{ field.size }}] = {0}{%- else %} = 0{% endif %};
{%- endfor %}


    const char* _json_string = R"({
{%- for field in fields %}
    "{{ field.token_name }}": {{ field.format_specifier }}{% if not loop.last -%},{% endif -%}
{%- endfor %}
})";

{%- for field in fields %}  
    const char* {{ field.internal_name | upper }}_TOKEN_NAME = "{{ field.token_name }}";
    const uint8_t {{ field.internal_name  | upper }}_TOKEN_ID = {{ field.token_id }};
{%- endfor %}
    const uint8_t _NUM_TOKENS  = {{ proto.num_tokens }};
};

#endif /* {{ package_name | upper }}_PROTO_H */
"""

class TitaniumFileGenerator:
    _SUPPORTED_TYPES = [
        "uint8_t",
        "int8_t",
        "uint16_t",
        "int16_t",
        "uint32_t",
        "int32_t",
        "uint64_t",
        "int64_t",
        "float",
        "double",
        "string",
    ]

    def __init__(self):
        """
        Initializes a new instance of the TitaniumProto class.
        """
        self._content = None
        self._package_name = None
        self._fields = []

                # Create a Jinja2 Template object
        self._template = Template(template_string)

        # # Render the template with the data and print or save to a file
        # rendered_content = template.render(template_data)

        # env = Environment(loader=FileSystemLoader("./titanium_proto/templates"))
        # self._template = env.get_template("protobuf_template.jinja2")

    def _read_file(self, filepath: str):
        """
        Read content from a JSON file.

        Args:
            filepath (str): Path to the Titanium Protobuf JSON file.
        """
        with open(filepath, "r") as file:
            self._content = json.load(file)

    def _validate_syntax(self):
        """
        Validates the syntax of the protocol file.

        Raises:
            ValueError: If the syntax is not "titanium1".
        """
        if self._content.get("syntax") != "titanium1":
            raise ValueError(
                "Invalid syntax: protocol file must have 'syntax' set to 'titanium1'."
            )

    def _update_package_name(self):
        """
        Updates the package name by parsing the protocol file content.

        Raises:
            ValueError: If the package name is missing.
        """
        self._package_name = self._content.get("package")
        if self._package_name is None:
            raise ValueError("Missing Package Name in protocol file.")

    def _parse_fields(self):
        """
        Parse and validate fields from the protocol file content.

        Raises:
            ValueError: If a field has an unsupported data type.
        """
        token_id = 1
        for field in self._content.get("fields"):
            field_type = field.get("type")
            field["token_id"] = token_id
            if field_type not in self._SUPPORTED_TYPES:
                supported_types_str = ", ".join(self._SUPPORTED_TYPES)
                raise ValueError(
                    f"Unsupported data type: {field_type}. Supported types are: {supported_types_str}."
                )

            self._fields.append(TitaniumField(field))
            token_id += 2

    def import_and_parse_proto_file(self, filepath: str):
        """
        Imports and parses definitions from a Titanium protobuf file.

        Args:
            filepath (str): Path to the Titanium protobuf file.
        
        Raises:
            ValueError: If there are syntax errors in the protobuf file or required fields are missing.
        """
        self._read_file(filepath)
        self._validate_syntax()
        self._update_package_name()
        self._parse_fields()
        
    def generate_header_file(self, redirect_outfile: str = ""):
        data = {}
        serialized_size_list = []
        maximum_size_list = []
        minimum_size_list = []
        static_maximum_size_list = []
        num_of_arrays = 0
        
        data["package_name"] = self._package_name
        data["fields"] = []
        for field in self._fields:
            data["fields"].append(field.to_dict())
            
            if field.is_array:
                serialized_size_list.append(f"(strlen(this->{field.internal_name}) + 1)")
                static_maximum_size_list.append(f"{field.defined_size}")
                num_of_arrays += 1
            else:
                serialized_size_list.append(f"sizeof(this->{field.internal_name})")
                static_maximum_size_list.append(f"sizeof({field.type_name})")
                minimum_size_list.append(f"sizeof(this->{field.internal_name})")
                
            maximum_size_list.append(f"sizeof(this->{field.internal_name})")

        data["proto"] = {}            
        data["proto"]["serialized_size"] =  " + ".join(serialized_size_list)
        data["proto"]["maximum_size"] = " + ".join(maximum_size_list)
        data["proto"]["minimum_size"] = " + ".join(minimum_size_list) + f" + {num_of_arrays}"
        data["proto"]["static_maximum_size"] = " + ".join(static_maximum_size_list)
        data["proto"]["num_tokens"] = (len(self._fields) * 2) + 1
        
        rendered_code = self._template.render(data)
        
        with open(f"{redirect_outfile}{self._package_name}Proto.h", 'w') as file:
            file.write(rendered_code)